/******************************************************************************
 expr.c
 ******************************************************************************
 
 Description:

 Let's assume the following expression: 2 + a + 5 % 2 * 3

 First, a tree is built as follows:
					     +
					    / \ 
					   +   *
					  /|  / \
					 2 a %   3
					    / \
					   5   2

High priority operations are placed low down in the tree, the lowest
priority operation is on top.

Now the optimization can begin. It is performed in two steps. The first
step is called Reduce(). It calculates all branches with constant
expressions. After Reduce() on the above tree, it will look like this:

					     +
					    / \ 
					   +   3
					  / \
					 2   a

Now it is time for the second step. It is called MergeConstants(). It will
try to move constants together to create branches that Reduce() can
remove. It starts from the top and moves constants down by swapping them
with the underlying expressions. After MergeConstants() the tree will look
like this:

					     +
					    / \ 
					   +   a
					  / \
					 2   3

Now we start the process over again with Reduce(). After Reduce() the tree
is even smaller:

					     +
					    / \ 
					   5   a

Now we are finished. The final expression will be 5 + a.

The optimizations that are performed are:

- If two operands are constants, they are combined.
- If a left operand is 0 and the operator is '&&' then the entire
  expression is set to 0.
- If a left operand is 1 and the operator is '||' then the entire
  expression is set to 1.
- Multiplications and divisions with 1 are eliminated.

There are other optimizations that can be done. When MergeConstants sees
that it is about to swap two constants, it will combine them if they have
the same operator.

----

Author: Daniel Stenberg <Daniel.Stenberg@sth.frontec.se>         

 ************************************************************************/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "types.h"
#include "lists.h"
#include "expr.h"
#ifdef DBASE
#include "dbase.h"
#endif

static EXPR_NODE *Build(EXPR_OPER_LIST *operlist, EXPR_NODE *left);

void debugf(char *fmt, ...)
{
#ifdef DEBUG
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stdout, fmt, ap);
  va_end(ap);
#endif  
}

/****************************************************************
** GetPrio()	- Get the priority of a type
**
**************/
static int GetPrio(EXPR_TYPE type)
{
   switch(type)
   {
      case COMP_DIVISION:
      case COMP_MULTIPLY:
      case COMP_REMAIN:
	 return(0);

      case COMP_PLUS:
	 return(1);

      case COMP_SHIFTLEFT:
      case COMP_SHIFTRIGHT:
	 return(2);

      case COMP_LESS:
      case COMP_LESSEQ:
      case COMP_GREATER:
      case COMP_GREATEQ:
	 return(3);

      case COMP_EQUAL:
      case COMP_NOTEQUAL:
	 return(4);

      case COMP_BINARYAND:
	 return(5);

      case COMP_XOR:
	 return(6);
	 
      case COMP_BINARYOR:
	 return(7);
	 
      case COMP_LOGICAND:
	 return(8);
	 
      case COMP_LOGICOR:
	 return(9);

      default:
	 return(100);	/* No priority (for COMP_NOTHING) */
   }
}

static EXPR_NODE *GetNextElement(EXPR_OPER_LIST *operlist, int step)
{
  EXPR_LIST_ENTRY *oper;

  oper = operlist->entryptr;
  
  if(step)
  {
    operlist->entryptr = (EXPR_LIST_ENTRY *)
      GetNext((LIST_NODE *)operlist->entryptr);
  }
  return(&oper->oper);
}

static BOOL EndOfExpression(EXPR_OPER_LIST *operlist)
{
  return(!operlist->entryptr || (COMP_LAST == operlist->entryptr->oper.type));
}

#ifdef DEBUG
static void PrintOperator(EXPR_TYPE oper)
{
  switch(oper)
  {
  case COMP_EQUAL:
    debugf(" == ");
    break;
    
  case COMP_LESS:
    debugf(" < ");
    break;
    
  case COMP_GREATER:
    debugf(" > ");
    break;
    
  case COMP_LESSEQ:
    debugf(" <= ");
    break;
    
  case COMP_GREATEQ:
    debugf(" >= ");
    break;
    
  case COMP_NOTEQUAL:
    debugf(" != ");
    break;
    
  case COMP_PLUS:
    debugf(" + ");
    break;
    
  case COMP_MINUS:
    debugf(" - ");
    break;
    
  case COMP_LOGICOR:
    debugf(" || ");
    break;
    
  case COMP_LOGICAND:
    debugf(" && ");
    break;
    
  case COMP_MULTIPLY:
    debugf(" * ");
    break;
    
  case COMP_DIVISION:
    debugf(" / ");
    break;
    
  case COMP_REMAIN:
    debugf(" % ");
    break;
    
  case COMP_BINARYOR:
    debugf(" | ");
    break;
    
  case COMP_BINARYAND:
    debugf(" & ");
    break;
    
  case COMP_XOR:
    debugf(" ^ ");
    break;
    
  case COMP_SHIFTLEFT:
    debugf(" << ");
    break;
    
  case COMP_SHIFTRIGHT:
    debugf(" >> ");
    break;
    
  default:
    break;
  }
}

#endif

void PrintType(EXPR_TYPE type)
{
#ifdef DEBUG
  switch(type)
  {
  case COMP_INTEGER:
    debugf("COMP_INTEGER");
    break;
  case COMP_STRING:
    debugf("COMP_STRING");
    break;
  default:
    break;
  }
#endif
}

static EXPR_NODE *GetNextOper(EXPR_OPER_LIST *operlist, EXPR_TYPE expected)
{
  EXPR_NODE *node;

  if(EndOfExpression(operlist))
  {
    return(NULL);
  }
  
  node = GetNextElement(operlist, 1);
  node->left = node->right = NULL;

  if((node->type > COMP_OPERATOR) && (expected < COMP_OPERATOR))
  {
    debugf("GetNextOper() error - value expected\n");
  }
  
  return(node);
}

EXPR_NODE *Build(EXPR_OPER_LIST *operlist, EXPR_NODE *oper)
{
  EXPR_NODE *right;
  EXPR_NODE *next_oper;
  int prio;
  int i;
  
  prio = GetPrio(oper->type);

  if(!(oper->right = GetNextOper(operlist, COMP_OPERAND)))
  {
    debugf("Build() Error - premature end of expression\n");
    exit(EXPR_SYNTAX_ERROR);
  }

  if(oper->right->type == COMP_LEFT_PAREN)
  {
    right = CreateTree(operlist, TRUE);
    for(i = 0;i < oper->right->num_unary;i++)
    {
      right->unary[i] = oper->right->unary[i];
    }
    right->num_unary = oper->right->num_unary;
    oper->right = right;
  }
  
  if((next_oper = GetNextOper(operlist, COMP_OPERATOR)))
  {
    if(next_oper->type != COMP_RIGHT_PAREN)
    {
      if(GetPrio(next_oper->type) >= prio)
      {
	next_oper->left = oper;
	oper = Build(operlist, next_oper);
      }
      else
      {
	next_oper->left = oper->right;
	right = Build(operlist, next_oper);
	oper->right = right;
      }
    }
  }
  return(oper);
}

EXPR_NODE *CreateTree(EXPR_OPER_LIST *operlist, BOOL subexpression)
{
  EXPR_NODE *left;
  EXPR_NODE *tmp;
  EXPR_NODE *oper;
  int i;

  /* First get the left operand, which can be either a value or a left
     parenthesis */
  if(!(left = GetNextOper(operlist, COMP_OPERAND)))
  {
    debugf("CreateTree() Error 1 - premature end of expression\n");
    exit(EXPR_SYNTAX_ERROR);
  }

  if(left->type == COMP_LEFT_PAREN)
  {
    tmp = CreateTree(operlist, TRUE);
    
    for(i = 0;i < left->num_unary;i++)
    {
      debugf("Transferring unary\n");
      tmp->unary[i] = left->unary[i];
    }
    tmp->num_unary = left->num_unary;
    left = tmp;
  }
  
  if(!(oper = GetNextOper(operlist, COMP_OPERATOR)))
  {
    if(subexpression)
    {
      debugf("CreateTree() Error 2 - premature end of expression\n");
      exit(EXPR_SYNTAX_ERROR);
    }
    else
    {
      return(left);
    }
  }
  
  if(oper->type == COMP_RIGHT_PAREN)
  {
    return(left);
  }
  oper->left = left;

  return(Build(operlist, oper));
}

#ifdef DEBUG
/***************************************************************
** OutputTree()	- Debugging function to output the complete
**		  expression tree.
**
******************/
void OutputTree(EXPR_NODE *tree)
{
  if (tree->type < COMP_OPERATOR)
  {
    switch(tree->type)
    {
    case COMP_INTEGER:
      debugf("%d", tree->num);
      break;
    case COMP_STRING:
      debugf("\"%s\"", tree->str);
      break;
    case COMP_FUNCTION:
      debugf("%s()", tree->name);
      break;
    case COMP_DATAFIELD:
      debugf("%s", tree->name);
      break;
    default:
      break;
    }
    return;
  }

  OutputTree(tree->left);
  PrintOperator(tree->type);
  OutputTree(tree->right);
}
#endif

#ifdef OPTIMIZE
/***************************************************************
** Calculate()	- This is the function that combines two nodes
**		  into one.
**
*****************/
static EXPR_RESULT Calculate(EXPR_TYPE type, long a, long b, long *result)
{
   switch (type)
   {
   case COMP_MULTIPLY:
     debugf("%d * %d, ", a, b);
     a *= b;
     break;

   case COMP_DIVISION:
     debugf("%d / %d, ", a, b);
     if(!b) {
       /* Division by zero!!! */
       /*       debugf("[Div by 0]"); */
       *result = 0;
       return EXPR_DIVISION_BY_ZERO;
     }
     a /= b;
     break;

   case COMP_REMAIN:
     debugf("%d % %d, ", a, b);
     if(!b) {
       /* Division by zero!!! */
       /* debugf("[Div by 0]"); */
       *result = 0;
       return EXPR_DIVISION_BY_ZERO;
     }
     a %= b;
     break;

   case COMP_PLUS:
     debugf("%d + %d, ", a, b);
     a += b;
     break;

   case COMP_MINUS:
     debugf("%d - %d, ", a, b);
     a -= b;
     break;

   case COMP_SHIFTLEFT:
     debugf("%d << %d, ", a, b);
     a <<= b;
     break;

   case COMP_SHIFTRIGHT:
     debugf("%d >> %d, ", a, b);
     a >>= b;
     break;

   case COMP_LESS:
     debugf("%d < %d, ", a, b);
     a = (a < b);
     break;

   case COMP_GREATER:
     debugf("%d > %d, ", a, b);
     a = (a > b);
     break;

   case COMP_LESSEQ:
     debugf("%d <= %d, ", a, b);
     a = (a <= b);
     break;

   case COMP_GREATEQ:
     debugf("%d >= %d, ", a, b);
     a = (a >= b);
     break;

   case COMP_EQUAL:
     debugf("%d == %d, ", a, b);
     a = (a == b);
     break;

   case COMP_NOTEQUAL:
     debugf("%d != %d, ", a, b);
     a = (a != b);
     break;

   case COMP_BINARYAND:
     debugf("%d & %d, ", a, b);
     a &= b;
     break;

   case COMP_XOR:
     debugf("%d ^ %d, ", a, b);
     a ^= b;
     break;

   case COMP_BINARYOR:
     debugf("%d | %d, ", a, b);
     a |= b;
     break;

   case COMP_LOGICAND:
     debugf("%d && %d, ", a, b);
     a = (a && b);
     break;

   case COMP_LOGICOR:
     debugf("%d || %d, ", a, b);
     a = (a || b);
     break;
   }
   *result = a;
   return EXPR_OK;
}
#endif

/***************************************************************
** CalculateTree() - This is the function that calculates
**		     a whole tree.
**
*****************/
EXPR_RESULT CalculateTree(EXPR_NODE *node, EXPR_NODE *result, void *data)
{
   EXPR_RESULT err;
   EXPR_NODE left;
   EXPR_NODE right;
   int i;     

   if(node->type > COMP_OPERATOR)
   {
     if((err = CalculateTree(node->left, &left, data)))
     {
       debugf("(Error left: %d)", err);
       return(err);
     }
      
     if((err = CalculateTree(node->right, &right, data)))
     {
       debugf("(Error right: %d)", err);
       return(err);
     }

     debugf("Types: ");
     PrintType(left.type);
     debugf(" and ");
     PrintType(right.type);
     debugf(": ");
     
     switch (node->type)
     {
     case COMP_MULTIPLY:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num * right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d * %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_DIVISION:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num * right.num;
	 if(!right.num) {
	   /* Division by zero!!! */
	   /*       debugf("[Div by 0]"); */
	   result->num = 0;
	   return EXPR_DIVISION_BY_ZERO;
	 }
	 result->num = left.num / right.num;
 	 result->type = COMP_INTEGER;
	 debugf("%d / %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_REMAIN:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 if(!right.num) {
	   /* Division by zero!!! */
	   /* debugf("[Div by 0]"); */
	   result->num = 0;
	   return EXPR_DIVISION_BY_ZERO;
	 }
	 result->num = left.num % right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d %% %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_PLUS:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num + right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d + %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_MINUS:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num - right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d - %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_SHIFTLEFT:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num << right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d << %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_SHIFTRIGHT:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num >> right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d >> %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_LESS:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num < right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d < %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_GREATER:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num > right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d > %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_LESSEQ:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num <= right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d <= %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_GREATEQ:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num >= right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d >= %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_EQUAL:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num == right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d == %d = %d\n", left.num, right.num, result->num);
       }
       else if(IS_STRING(left.type) && IS_STRING(right.type))
       {
	 result->num = !strcmp(left.str, right.str)?TRUE:FALSE;
	 result->type = COMP_INTEGER;
	 debugf("\"%s\" == \"%s\" = %d\n", left.str, right.str, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_NOTEQUAL:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num != right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d != %d = %d\n", left.num, right.num, result->num);
       }
       else if(IS_STRING(left.type) && IS_STRING(right.type))
       {
	 result->num = strcmp(left.str, right.str)?TRUE:FALSE;
	 result->type = COMP_INTEGER;
	 debugf("\"%s\" == \"%s\" = %d\n", left.str, right.str, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_BINARYAND:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num & right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d & %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_XOR:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num ^ right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d ^ %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_BINARYOR:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = left.num | right.num;
	 result->type = COMP_INTEGER;
	 debugf("%d | %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_LOGICAND:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num && right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d && %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;
       
     case COMP_LOGICOR:
       if(IS_INTEGER(left.type) && IS_INTEGER(right.type))
       {
	 result->num = (left.num || right.num);
	 result->type = COMP_INTEGER;
	 debugf("%d || %d = %d\n", left.num, right.num, result->num);
       }
       else
       {
	 return(EXPR_INVALID_TYPE);
       }
       break;

     default:
       debugf("FAN! %d", node->type);
       break;
     }
   }
   else
   {
     if(node->type == COMP_FUNCTION)
     {
       debugf("COMP_FUNCTION ");
       err = node->func->func(node, result, data);
       if(err)
	 return(err);
     }
#ifdef DBASE
     else if(node->type == COMP_DATAFIELD)
     {
       t_field *f;

       if((f = getfield(data, node->name)))
       {
	 if(f->template->type == string)
	 {
	   debugf("COMP_DATAFIELD(\"%s\")\n", f->data);
	   result->type = COMP_STRING;
	   result->str = f->data;
	 }
	 else if(f->template->type == integer)
	 {
	   debugf("COMP_DATAFIELD(%d)\n", f->value);
	   result->type = COMP_INTEGER;
	   result->num = f->value;
	 }
	 else
	 {
	   return EXPR_INVALID_TYPE;
	 }
       }
     }
#endif
     else if(IS_INTEGER(node->type))
     {
       debugf("COMP_INTEGER(%d)\n", node->num);
       result->num = node->num;
       result->type = node->type;
     }
     else if(IS_STRING(node->type))
     {
       debugf("COMP_STRING(\"%s\")\n", node->str);
       result->str = node->str;
       result->type = node->type;
     }
   }
   
   for(i = 0;i < node->num_unary;i++)
   {
     switch(node->unary[i])
     {
     case COMP_NOTOPERATOR:
       result->num = !result->num;
       break;
     case COMP_NEGATE:
       result->num = ~result->num;
       break;
     case COMP_MINUS:
       result->num = -result->num;
       break;
     case COMP_PLUS:
       result->num = +result->num;
       break;
     default:
       break;
     }
   }
	 
   return EXPR_OK;
}

#ifdef OPTIMIZE
/*******************************************************************
** Reducable()	- Tells if a certain expression is reducable. It
**		  returns a code telling which optimization that can
**		  be done.
**
**********************/
static EXPR_OPT_TYPE Reducable(EXPR_TYPE type, EXPR_NODE *left, EXPR_NODE *right)
{
   /* If the two operands are constant, then just combine them */
  if((left->type==COMP_INTEGER) && (right->type==COMP_INTEGER))
    return(OPT_TWO_CONSTANTS);

  return(OPT_IMPOSSIBLE);
}

/*********************************************************************
** Reduce()	- Searches the tree recursively for constant expressions
**		  to combine. All branches with a left and a right node
**		  that can be reduced are combined.
**		  If any changes was made, the 'changes' variable will be
**		  non-zero.
**
***************************/
static EXPR_RESULT Reduce(struct PNode *node, struct PNode **new)
{
   struct PNode *left, *right;
   EXPR_RESULT err;
   long result;

   /* Type of optimization that can be made */
   EXPR_OPT_TYPE opt_type = OPT_IMPOSSIBLE;

   /* Return if the node is already an operand */
   if (node->data.type < COMP_OPERATOR) {
     *new = node;
     return EXPR_OK;
   }

   /* Search the tree recursively to the left and the right */
   if(err = Reduce(node->left, &left))
     return err;
   if(err = Reduce(node->right, &right))
     return err;

   /* Find out if the branch can be combined */
   opt_type = Reducable(node->data.type, &left->data, &right->data);

   /* Perform the proper optimizations */
   switch (opt_type) {
   case OPT_TWO_CONSTANTS:
     /* If both operands are constants, then just combine them */
     err = Calculate(node->data.type,
                     left->data.num,
                     right->data.num,
                     &result);    
     left->data.num = result;
     node->data = left->data;
     node->left = NULL;
     node->right = NULL;
     node->data.type = COMP_INTEGER;
     break;

      /* If no optimization could be made, then just return */	 
   case OPT_IMPOSSIBLE:
     break;
   }
   *new = node;
   return err;
}

/*****************************************************************
** OptimizeTree	- Optimize the entire expression tree. Reduce()
**		  and MergeConstants() are called until no changes
**		  are made.
**
***************************/
static EXPR_RESULT OptimizeTree(EXPR_OPER_LIST *operlist, struct PNode *top)
{
  int pass_number = 1;
  struct PNode *new;
  EXPR_RESULT err=EXPR_OK;

#ifdef DEBUG   
  debugf("------------------------------------------------------------------\n");
  debugf("Pass %d:\t\t", pass_number++);
  OutputTree(top);
  debugf("\nReducing:\t");
#endif
  err = Reduce(top, &new); /* this new one isn't used since it is reduced to
                              zero */
  if(err)	/* Break if an error has occurred */
    return err;

#ifdef DEBUG
  debugf("\nReduced:\t");
  OutputTree(top);
  debugf("\n");
#endif
  return err;
}

/*****************************************************************
** OptimizeExpression()	- Optimize an entire expression tree.
**
***************************/
static EXPR_RESULT OptimizeExpression(EXPR_OPER_LIST *operlist, long *result)
{
   struct PNode *top;	/* The tree root */
   EXPR_NODE oper;
   EXPR_RESULT ret;
   
   top = ParseExpression(operlist);

   ret = OptimizeTree(operlist, top);

   *result =  top->data.num;

   /* PutOper(&oper,ReturnExpression(top, TRUE)); */
   return ret;
}
#endif

EXPR_RESULT ParseExpression(char *in,      /* pointer to the expression */
                   char **done,   /* pointer to the pointer that will be set
                                     by the function to where the expression
                                     ended. May be NULL if the information
                                     isn't interesting. */
                   EXPR_VARIABLE *vars, /* pointer to a list of variable
                                             names and their 'values' for
                                             variable substitution. May be
                                             NULL if no variables is
                                             available. */
                   EXPR_FUNCTION *funcs, /* pointer to a list of function
                                             names and their parameter counts
					     */
		   EXPR_OPER_LIST *exprList)	/* The list to append the nodes to */
{
  char *rest;
  long num;
  EXPR_RESULT ret = EXPR_OK;
  BOOL negative = FALSE;
  BOOL operand = TRUE; /* take turns, operand/operator, that's how expressions
                          work. =) */
  BOOL cont=TRUE;
  EXPR_LIST_ENTRY *newentry;
  EXPR_NODE expr;

  exprList->entryptr = (EXPR_LIST_ENTRY *)GetFirst(&exprList->list);

  while(*in && cont && (ret == EXPR_OK)) {
    expr.type = COMP_LAST;
    expr.num = 0;
    expr.num_unary = 0;

    while(isspace(*in))
      in++;

    if(!*in)
      break;

    if(!operand) {
      char check = *in++;
      switch(check) {
      case '*':
        expr.type = COMP_MULTIPLY;
	debugf(" * ");
        break;
      case '/':
        expr.type = COMP_DIVISION;
	debugf(" / ");
        break;
      case '%':
        expr.type = COMP_REMAIN;
	debugf(" % ");
        break;
      case '+':
        expr.type = COMP_PLUS;
	debugf(" + ");
        break;
      case '-':
        expr.type = COMP_MINUS;
	debugf(" - ");
        break;
      case '<':
        /* this could be three different ones, check the following letter */
        switch(*in++) {
        case '<':
          expr.type = COMP_SHIFTLEFT;
	  debugf(" << ");
          break;
        case '=':
          expr.type = COMP_LESSEQ;
	  debugf(" <= ");
          break;
        default:
          in--;
          expr.type = COMP_LESS;
	  debugf(" < ");
          break;
        }
        break;
      case '>':
        /* this could be three different ones, check the following letter */
        switch(*in++) {
        case '>':
          expr.type = COMP_SHIFTRIGHT;
	  debugf(" >> ");
          break;
        case '=':
          expr.type = COMP_GREATEQ;
	  debugf(" >= ");
          break;
        default:
          in--;
          expr.type = COMP_GREATER;
	  debugf(" > ");
	  break;
        }
        break;
      case '=':
        /* we do support BOTH == and = for equality checks since we don't
           do assigns anyway */
        if('=' == *in)
          in++;
        expr.type = COMP_EQUAL;
	debugf(" == ");
        break;
      case '!':
        if('=' != *in) {
          ret = EXPR_ILLEGAL_OPERATOR;
          continue;
        }
        in++;
        expr.type = COMP_NOTEQUAL;
	debugf(" != ");
        break;
      case '&':
        /* could be two different ones */
        if('&' != *in)
	{
          expr.type = COMP_BINARYAND;
	  debugf(" & ");
	}
        else {
          expr.type = COMP_LOGICAND;
          in++;
	  debugf(" && ");
        }
        break;
      case '|':
        /* could be two different ones */
        if('|' != *in)
	{
          expr.type = COMP_BINARYOR;
	  debugf(" | ");
	}
        else {
          expr.type = COMP_LOGICOR;
          in++;
	  debugf(" || ");
        }
        break;

      case '^':
        expr.type = COMP_XOR;
	debugf(" ^ ");
        break;

      default:
        in--; /* get back onto the "operator" */
#ifdef OPERATOR_ALIASES
        if(isalpha(*in)) {
          size_t len =1;
          int i;
          char *ptr = in;
          static EXPR_OPER_ALIAS aliases[]={
            {"and", COMP_LOGICAND},
            {"or", COMP_LOGICOR},
            {"xor", COMP_XOR},
          };
          while(isalpha(*++in))
            len++;
          for(i=0; i<sizeof(aliases)/sizeof(aliases[0]); i++) {
            if((len ==strlen(aliases[i].alias)) &&
#ifdef WIN32
               !strnicmp(aliases[i].alias, ptr, len)) {
#else
               !strncasecmp(aliases[i].alias, ptr, len)) {
#endif
              expr.type = aliases[i].type;
              i = -1;
              break;
            }
          }
          if(i<0)
            break;
          else
            ret = EXPR_ILLEGAL_OPERATOR;
        }
#endif
        cont = FALSE; /* we abort on weird operators */
        continue;
      }
    }
    else {
      /* here comes an operand */
      BOOL not=FALSE;
      BOOL negate=FALSE;
      BOOL minus=FALSE;
      BOOL plus=FALSE;
      char *unary=in; /* here comes a flow of unary operators */
      char *endofu;
      
      /* Eat ! and ~ here */
      while(isspace(*in) ||
            ('!' == *in) ||
            ('~' == *in) ||
            ('+' == *in) ||
            ('-' == *in)) {
        if('!' == *in)
          not = TRUE;
        else if('~' == *in)
          negate = TRUE;
        else if('-' == *in)
          minus = TRUE;
        else if('+' == *in)
          plus = TRUE;
        in++;
      }
      if(not || negate || minus || plus)
        endofu = in; /* set the end-of-unary section */

      /* An operand comes up.
         This can be any item of:
         - number (decimal, octal, hexadecimal)
         - variable (a sequence of letters)
         - an open parenthesis that starts a sub-expression
         */

      /* try to get a number: */
      num = strtol(in, &rest, 0); /* the 0 here make us support octal and
                                     hex with no fuss */
      if(in == rest) {
        /* resolve this variable to a number, or return error, we should NOT
           keep any variables in the expression */
        size_t len=0;
        if('(' == *in) {
          in++;
	  debugf("(");
	  newentry = (EXPR_LIST_ENTRY *)malloc(sizeof(EXPR_LIST_ENTRY));
	  debugf("malloc: (entry) 0x%lx\n", newentry);
	  if(newentry) {
	    newentry->oper = expr;
	    newentry->oper.type = COMP_LEFT_PAREN;
	    if(not || negate || minus || plus) {
	      /* Eat ! and ~ here, from right-to-left */
	      char *ptr=--endofu;
	      while((isspace(*ptr) ||
		     ('!' == *ptr) ||
		     ('~' == *ptr) ||
		     ('+' == *ptr) ||
		     ('-' == *ptr)) && (ptr>=unary)) {
                if(MAX_UNARY == newentry->oper.num_unary) {
                  /* BAAAAD expression uses too many unary operators on a
                     single operand! */
                  return EXPR_TOO_MANY_UNARY;
                }
		if('!' == *ptr)
		  newentry->oper.unary[newentry->oper.num_unary++] = COMP_NOTOPERATOR;
		else if('~' == *ptr)
		  newentry->oper.unary[newentry->oper.num_unary++] = COMP_NEGATE;
		else if('-' == *ptr)
		  newentry->oper.unary[newentry->oper.num_unary++] = COMP_MINUS;
		else if('+' == *ptr)
		  newentry->oper.unary[newentry->oper.num_unary++] = COMP_PLUS;

		debugf("Unary before paren\n");
		ptr--;
	      }
	    }

	    if(!AddLast(&exprList->list, &newentry->list_node))
	      return EXPR_OUT_OF_MEMORY;
	  }
	  else
	    return EXPR_OUT_OF_MEMORY;

          ret = ParseExpression(in, &in, vars, funcs, exprList);
          if(ret)
            break;
          else if(')' != *in) {
            ret = EXPR_NO_END_OF_PARENTHESIS;
            break;
          }
          else
	  {   
            in++;
	    debugf(")");
	    expr.type = COMP_RIGHT_PAREN;
	  }
        }
	else if('"' == *in)
	{
	  char *p;
	  unsigned long len = 0;
	  
	  p = expr.str = malloc(STRINGSIZE);
	  debugf("malloc: (string) 0x%lx\n", p);
	  if(!p)
	  {
	    return EXPR_OUT_OF_MEMORY;
	  }
	  in++;
	  while(*in && ('"' != *in) && (len < 256))
	  {
	    if(('\\' == *in) && ('"' == *(in+1)))
	    {
	      *p++ = '"';
	      in+=2;
	      len+=2;
	    }
	    else
	    {
	      *p++ = *in++;
	      len++;
	    }
	  }
	  if(len > STRINGSIZE-1)
	  {
	    return EXPR_STRING_TOO_LONG;
	  }
	  else if(!*in)
	  {
	    return EXPR_STRING_NOT_TERMINATED;
	  }
	  
	  *p = '\0';
	  in++;
	  
	  expr.type = COMP_STRING;
	  debugf("\"%s\"", expr.str);
	}
        else {
#ifndef DBASE
          EXPR_VARIABLE *thisvar=vars;
#endif
          EXPR_FUNCTION *thisfunc=funcs;
          char *inname=in;
          BOOL found=FALSE;
          while(isalnum(*in)) {
            in++;
            len++;
          }
          if(!len) {
            ret = EXPR_SYNTAX_ERROR; /* no alpha */
            break;
          }
          /* pass the spaces */
          while(isspace(*in))
            in++;
          if('(' == *in) {
            /* This is a FUNCTION! */
            in++; /* pass it */

            if(thisfunc) {
              while(thisfunc->name) {
                if((len == strlen(thisfunc->name)) &&
                   !strncmp(thisfunc->name, inname, len)) {
                  /* this is the variable we want */
		  strncpy(expr.name, inname, len);
		  expr.name[len] = '\0';
		  debugf("f_%s(", expr.name);
		  expr.type = COMP_FUNCTION;
		  expr.func = thisfunc;
                  thisfunc->flags |= FUNC_ACCESSED; /* this was read */
                  found = TRUE;
                  break;
                }
                thisfunc++;
              }
            }
            if(!found) {
              ret = EXPR_NO_SUCH_FUNCTION;
              break;
            }

	    expr.num_params = 0;

	    /* deal with parameters */
	    while(!ret && (expr.num_params < expr.func->num_params))
	    {
	      /* pass the spaces */
	      while(isspace(*in))
		in++;
	      if(')' == *in) {
		in++;
		debugf(")");
		if((expr.num_params + 1) < expr.func->num_params)
		  return(EXPR_TOO_FEW_PARAMETERS);
		break;
	      }
	      /* pass the spaces */
	      while(isspace(*in))
		in++;
	        InitList(&expr.params[expr.num_params].list, NO_SIZE_CHECK);
		ret = ParseExpression(in, &in, vars, funcs,
			      &expr.params[expr.num_params]);
		if(ret)
		  break;
		expr.params[expr.num_params].entryptr =
		  (EXPR_LIST_ENTRY *)GetFirst(&expr.params[expr.num_params].list);
		expr.par[expr.num_params] =
		  CreateTree(&expr.params[expr.num_params], FALSE);
#ifdef DEBUG
		OutputTree(expr.par[expr.num_params]);
#endif
	      /* pass the spaces */
	      while(isspace(*in))
		in++;
	      if(')' == *in) {
		in++;
		debugf(")");
		if((expr.num_params + 1) < expr.func->num_params)
		  return(EXPR_TOO_FEW_PARAMETERS);
		break;
	      }
	      else if(',' != *in)
	      {
		ret = EXPR_NO_END_OF_PARENTHESIS;
		break;
	      }
	      debugf(", ");
	      in++; /* Skip the comma */
	      expr.num_params++;
	    }
	    if(ret)
	      break;
	    if(expr.num_params > 8)
	    {
	      ret = EXPR_TOO_MANY_PARAMETERS;
	      break;
	    }
	    
          }
          else {
#ifdef DBASE
	    extern LIST dbTemplate;
	    t_fieldTemplate *f;
	    char n[256];
	    char *tmp = inname;

	    len=0;
	    while(isalnum(*tmp++))
	      len++;
	    strncpy(n, inname, len);
	    n[len] = 0;
	    debugf("Looking for %s\n", inname);
	    for ( f = (t_fieldTemplate*) GetFirst( &dbTemplate ); f;
		  f = (t_fieldTemplate*) GetNext( &f->node ) )
	    {
	      debugf("Trying %s\n", f->name);
	      if ( !strcmp( n, f->name ) )
	      {
		debugf("Found fieldname: %s\n", f->name);
		strcpy(expr.name, f->name);
		expr.type = COMP_DATAFIELD;
		found = TRUE;
		break;
	      }
	    }

	    if(!found)
	    {
	      ret = EXPR_NO_SUCH_DATA_FIELD;
	      break;
	    }
#else
            if(thisvar) {
              while(thisvar->name) {
                if((len == strlen(thisvar->name)) &&
                   !strncmp(thisvar->name, inname, len)) {
                  /* this is the variable we want */
		  strncpy(expr.name, inname, len);
		  expr.name[len] = '\0';
		  debugf("v_%s", expr.name);
		  expr.type = COMP_VARIABLE;
                  thisvar->flags |= VAR_ACCESSED; /* this was read */
                  found = TRUE;
                  break;
                }
                thisvar++;
              }
            }
            if(!found) {
              ret = EXPR_NO_SUCH_VARIABLE;
              break;
            }
#endif
          }
        }
      }
      else {
        expr.num = num;
	expr.type = COMP_INTEGER;
        in = rest;
      }
      if(not || negate || minus || plus) {
        /* Eat ! and ~ here, from right-to-left */
        char *ptr=--endofu;
        while((isspace(*ptr) ||
               ('!' == *ptr) ||
               ('~' == *ptr) ||
               ('+' == *ptr) ||
               ('-' == *ptr)) && (ptr>=unary)) {
          if(MAX_UNARY == expr.num_unary) {
            /* BAAAAD expression uses too many unary operators on a single
               operand! */
            return EXPR_TOO_MANY_UNARY;
          }
          if('!' == *ptr)
	    expr.unary[expr.num_unary++] = COMP_NOTOPERATOR;
          else if('~' == *ptr)
	    expr.unary[expr.num_unary++] = COMP_NEGATE;
          else if('-' == *ptr)
	    expr.unary[expr.num_unary++] = COMP_MINUS;
          else if('+' == *ptr)
	    expr.unary[expr.num_unary++] = COMP_PLUS;
          ptr--;
        }
      }

      if(negative) {
        expr.num = -expr.num;
        negative = FALSE;
      }
      if(expr.type == COMP_INTEGER)
      {
	debugf("%d", expr.num);
      }
    }
    newentry = (EXPR_LIST_ENTRY *)malloc(sizeof(EXPR_LIST_ENTRY));
    debugf("malloc: (entry) 0x%lx\n", newentry);
    if(newentry)
    {
      int i;
      newentry->oper = expr;
      if(expr.type == COMP_FUNCTION)
      {
	for(i = 0;i <= expr.num_params;i++)
	{
	  LIST_NODE *ln;
	  
	  InitList(&newentry->oper.params[i].list, NO_SIZE_CHECK);
	  for(ln = RemFirst(&expr.params[i].list);ln;ln = RemFirst(&expr.params[i].list))
	  {
	    AddLast(&newentry->oper.params[i].list, ln);
	  }
	}
      }
      if(!AddLast(&exprList->list, &newentry->list_node))
        return EXPR_OUT_OF_MEMORY;
    }
    else
      return EXPR_OUT_OF_MEMORY;

    operand ^= 1; /* toggle operand/operator flag */

  }
  if(EXPR_OK != ret)
    ;
  else if(expr.type != COMP_LAST && expr.type > COMP_OPERATOR) {
    ret = EXPR_SYNTAX_ERROR;
  }
  else {
#ifdef APA
    switch(OptimizeExpression(exprList, result)) {
    case EXPR_DIVISION_BY_ZERO:
      /*      debugf("Error - Division by zero!\n"); */
      ret = EXPR_DIVISION_BY_ZERO;
      break;
	 
    default:
      break;
    }
#endif
  }

  if(done)
    *done = in;

  exprList->entryptr = (EXPR_LIST_ENTRY *)GetFirst(&exprList->list);
  return ret;
}

/*******************************************************************
** DeleteExpression()	- Deletes all nodes in an expression list
**
**********************/
void DeleteExpression(EXPR_OPER_LIST *list)
{
  LIST_NODE *node;
  EXPR_LIST_ENTRY *ln;
  int i;

  for(node = RemFirst(&list->list);node;node = RemFirst(&list->list))
  {
    ln = (EXPR_LIST_ENTRY *)node;

    if(ln->oper.type == COMP_STRING)
    {
      debugf("free: (string) 0x%lx\n", ln->oper.str);
      free(ln->oper.str);
    }
    else if(ln->oper.type == COMP_FUNCTION)
    {
      debugf("Function (num_params = %d) - ", ln->oper.num_params+1);
      for(i = 0;i <= ln->oper.num_params;i++)
      {
	DeleteExpression(&ln->oper.params[i]);
      }
    }
    debugf("free: (entry) 0x%lx\n", node);
    free(node);
  }
}

static char *error_table[] =
{
  "", /* EXPR_OK */
  "Division by zero", /* EXPR_DIVISION_BY_ZERO */
  "Illegal operator", /* EXPR_ILLEGAL_OPERATOR */
  "No closing parenthesis", /* EXPR_NO_END_OF_PARENTHESIS */
  "Too few function parameters", /* EXPR_TOO_FEW_PARAMETERS */
  "Too many function parameters", /* EXPR_TOO_MANY_PARAMETERS */
  "Syntax error", /* EXPR_SYNTAX_ERROR */
  "Out of memory", /* EXPR_OUT_OF_MEMORY */
  "No such variable", /* EXPR_NO_SUCH_VARIABLE */
  "No such data field", /* EXPR_NO_SUCH_DATA_FIELD */
  "No such function", /* EXPR_NO_SUCH_FUNCTION */
  "Invalid type", /* EXPR_INVALID_TYPE */
  "String too long", /* EXPR_STRING_TOO_LONG */
  "String not terminated", /* EXPR_STRING_NOT_TERMINATED */
  "Too many unary operators" /* EXPR_TOO_MANY_UNARY */
};

/*******************************************************************
** ExprErrorText()	- Returns the error text corresponding to an
**			  error code
**
**********************/
char *ExprErrorText(EXPR_RESULT code)
{
  return error_table[code];
}
