#ifndef __OPTIMIZE_H
#define __OPTIMIZE_H

/************************************************************************
 expr.h

 External definitions for the expression parser.

 ----
 Features, functions and limits
 ==============================

 Full support for all C-style operators. Which means *, /, %, +, -, <, <<, <=,
 >, >>, >=, ==, !=, &, &&, |, || and ^. Additionally, it supports the unary
 operators +, -, ! and ~.

 The priority of the above operators are the very same as in the C programming
 language.

 Full support for "parenthesisized" sub-expressions.

 Operands can be decimal, octal (prefix with 0) or hexadecimal (prefix with
 '0x') numbers, any of a set of variables or an expression within parentheses.

 NOTE: Using '=' instead of '==' for equality is supported.

 ----
      Author: Daniel Stenberg <Daniel.Stenberg@sth.frontec.se>
 ************************************************************************/

/*
 * Operator enum:
 */

typedef enum {
  COMP_NOTHING, /* First, does NOTHING */
  COMP_OPERAND,
  COMP_INTEGER,
  COMP_STRING,
  COMP_DATAFIELD,
  COMP_VARIABLE,
  COMP_FUNCTION,
  COMP_LEFT_PAREN,
  COMP_RIGHT_PAREN,
  COMP_OPERATOR,	/* User to determine if operator or operand */
  COMP_ONESCOMPLEMENT,
  COMP_NOTOPERATOR,
  COMP_NEGATE,

  COMP_COMMA,
  COMP_NOTEQUAL,
  COMP_GREATER,
  COMP_SHIFTRIGHT,
  COMP_GREATEQ,
  COMP_LESS,
  COMP_SHIFTLEFT,
  COMP_LESSEQ,
  COMP_XOR,
  COMP_REMAIN,
  COMP_DIVISION,
  COMP_MULTIPLY,
  COMP_CONDOPEND,
  COMP_CONDOPSTART,
  COMP_MINUS,
  COMP_PLUS,
  COMP_BINARYOR,
  COMP_BINARYAND,
  COMP_LOGICOR,
  COMP_LOGICAND,
  COMP_EQUAL,

  COMP_LAST     /* Last, does NOTHING */
} EXPR_TYPE;

/*********************************************************
 * Return codes from ParseExpr()
 *********************************************************/
typedef enum
{
   EXPR_OK = 0,
   EXPR_DIVISION_BY_ZERO,
   EXPR_ILLEGAL_OPERATOR,
   EXPR_NO_END_OF_PARENTHESIS,
   EXPR_TOO_FEW_PARAMETERS,
   EXPR_TOO_MANY_PARAMETERS,
   EXPR_SYNTAX_ERROR,
   EXPR_OUT_OF_MEMORY,
   EXPR_NO_SUCH_VARIABLE,
   EXPR_NO_SUCH_DATA_FIELD,
   EXPR_NO_SUCH_FUNCTION,
   EXPR_INVALID_TYPE,
   EXPR_STRING_TOO_LONG,
   EXPR_STRING_NOT_TERMINATED,

   EXPR_TOO_MANY_UNARY,

   EXPR_LASTENTRY /* ALWAYS the _LAST_ */
} EXPR_RESULT;

#define STRINGSIZE 256

typedef enum
{
   OPT_IMPOSSIBLE = 0,
   OPT_TWO_CONSTANTS,		/* Two constants can be combined */
   OPT_AND_IS_FALSE,		/* The left operand of '&&' is FALSE */
   OPT_OR_IS_TRUE,		/* The left operand of '||' is TRUE */
   OPT_MULTIPLY_WITH_1,		/* A multiplication with 1 */
   OPT_DIVISION_WITH_1		/* A division with 1 */
} EXPR_OPT_TYPE;

#define OPERATOR_ALIASES

#ifdef OPERATOR_ALIASES
/* This is for operator aliases */

typedef struct
{
  char *alias;
  EXPR_TYPE type;
} EXPR_OPER_ALIAS;
#endif

typedef struct
{
  char *name; /* Make this unique and it will work! Make sure the last
                 name in the list is NULL to terminate it.*/
  long num;
#define VAR_NOTHING 0
#define VAR_ACCESSED 1 /* set by ParseExpr() when the variable has been
                          accessed */
  long flags;
} EXPR_VARIABLE;

typedef struct expr_oper_list
{
  LIST list;
  struct expr_list_entry *entryptr;
} EXPR_OPER_LIST;

/*********************************************************
** This is the operand/operator information structure
**
************************/
#define MAX_UNARY 8 /* maximum number of unary operators */
typedef struct expr_node
{
  struct expr_node *left;
  struct expr_node *right;
  EXPR_TYPE type;	/* The data type (<COMP_OPERATOR if an operand */
  int num_unary;	/* Number or unary operators */
  EXPR_TYPE unary[MAX_UNARY]; /* Unary operators to apply in order */
  char name[80];	/* Data field name if a field in the database
			   or a function */
  struct expr_function *func; /* Function definition */
  long num;		/* The constant numerical value if an operand */
  char *str;		/* String data if a string constant */
  int num_params;
  struct expr_oper_list params[8];	/* Function parameters if a function */
  struct expr_node *par[8];	/* Function parameters if a function */
} EXPR_NODE;

typedef struct expr_list_entry
{
  LIST_NODE list_node;
  EXPR_NODE oper;
} EXPR_LIST_ENTRY;

void debugf(char *fmt, ...);

typedef struct expr_function
{
  char *name; /* Make this unique and it will work! Make sure the last
                 name in the list is NULL to terminate it.*/
  EXPR_RESULT (*func)(struct expr_node *def, struct expr_node *res, void *data);
  EXPR_TYPE return_type;
  int num_params;
#define FUNC_NOTHING 0
#define FUNC_ACCESSED 1 /* set by ParseExpr() when the function has been
                          accessed */
  long flags;
} EXPR_FUNCTION;

#define IS_INTEGER(x) ((x) == COMP_INTEGER)
#define IS_STRING(x) ((x) == COMP_STRING)

/****************************************************************************
 * NAME
 *   ParseExpr()
 *
 * DESCRIPTION
 *   The expression parser function. Evaluates a C-style expression.
 *
 * RETURNS
 *   Error code. See the EXPR_RESULT enum above.
 *
 * INPUTS
 *   in     - pointer to the expression
 *   done   - pointer to the pointer that will be set by the function to where
 *            the expression parsing ended. May be NULL if the information
 *            isn't interesting.
 *
 ***************************************************************************/

EXPR_RESULT ParseExpression(char *in, char **done, EXPR_VARIABLE *vars,
		   EXPR_FUNCTION *funcs, EXPR_OPER_LIST *exprList);

EXPR_NODE *CreateTree(EXPR_OPER_LIST *operlist, BOOL subexpression);

EXPR_RESULT CalculateTree(EXPR_NODE *node, EXPR_NODE *result, void *data);

void DeleteExpression(EXPR_OPER_LIST *list);

char *ExprErrorText(EXPR_RESULT code);

#ifdef DEBUG
void OutputTree(EXPR_NODE *tree);
#endif
#endif
